import os
import sys
import logging
import joblib
import threading
import time
from datetime import datetime, timedelta
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
import requests
import json
import tensorflow as tf
from tensorflow.keras.models import Sequential, load_model
from tensorflow.keras.layers import LSTM, Dense, Dropout

# 프로젝트 루트 디렉터리와 관련된 모든 스크립트 디렉터리를 sys.path에 추가합니다.
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
scripts_dir = os.path.join(project_root, 'scripts')
config_dir = os.path.join(project_root, 'config')
data_dir = os.path.join(project_root, 'data')
utils_dir = os.path.join(scripts_dir, 'utils')

sys.path.extend([
    scripts_dir,
    os.path.join(scripts_dir, 'data_collection'),
    os.path.join(scripts_dir, 'data_processing'),
    os.path.join(scripts_dir, 'model'),
    os.path.join(scripts_dir, 'trading'),
    utils_dir
])

sys.path.append(config_dir)

# 주식 데이터 및 전체 데이터 모델 경로 설정
stock_model_path = os.path.join(data_dir, 'models', 'lstm_model_stock.h5')
stock_model_json_path = os.path.join(data_dir, 'models', 'lstm_model_stock.json')
stock_scaler_path = os.path.join(data_dir, 'models', 'scaler_stock.pkl')

all_model_path = os.path.join(data_dir, 'models', 'lstm_model_all.h5')
all_model_json_path = os.path.join(data_dir, 'models', 'lstm_model_all.json')
all_scaler_path = os.path.join(data_dir, 'models', 'scaler_all.pkl')

# 정확한 경로에서 필요한 함수들 임포트
from model_training import train_lstm_model, save_model, load_model, update_lstm_model
from datetime import datetime, time as dt_time
from stock_data_collector import fetch_data_for_detected_stocks, save_stock_codes_to_csv
from live_data_processing import real_time_prediction, update_real_time_model, get_stock_codes_from_csv
from trading_logic import execute_trades
from data_preprocessing import preprocess_data, load_csv_files_from_directory, normalize_data
from feature_engineering import add_features
from settings import settings  # settings 불러오기
available_cash = 0
seq = 0

def update_balance():
    # update_balance logic here, if necessary
    pass

def schedule_tasks():
    schedule.every(1).minute.do(update_balance)
    while True:
        schedule.run_pending()
        time.sleep(1)

def initialize():
    logging.info("Initializing system...")

def run_main():
    # 메인 실행 함수
    print("Running main function of main_total module...")
    # 여기에 메인 기능을 구현합니다
    # 예: 주요 작업 실행, 데이터 처리, 모델 예측 등

    # 예시: 필요한 모든 변수를 정의한 후에 main_total 모듈을 불러옵니다.
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
data_dir = os.path.join(project_root, 'data')

model_path = os.path.join(data_dir, 'models', 'lstm_model.h5')
model_json_path = os.path.join(data_dir, 'models', 'lstm_model.json')  # model_json_path 정의
scaler_path = os.path.join(data_dir, 'models', 'scaler.pkl')


import main_total

main_total.initialize()
main_total.run_main()


def is_data_modified(data_dir, last_training_time):
    """데이터 디렉토리 내의 파일 중 마지막 수정 시간이 주어진 시간보다 이후인지 확인."""
    for root, dirs, files in os.walk(data_dir):
        for file in files:
            file_path = os.path.join(root, file)
            if os.path.getmtime(file_path) > last_training_time.timestamp():
                return True
    return False

def should_train_model(data_dir, model_path):
    """모델이 다시 학습되어야 하는지 여부를 결정."""
    if not os.path.exists(model_path):
        return True  # 모델 파일이 없으면 학습해야 함

    last_training_time = datetime.fromtimestamp(os.path.getmtime(model_path))

    # 데이터 디렉토리에서 데이터가 변경되었는지 확인
    return is_data_modified(data_dir, last_training_time)

# 토큰 관련 함수들
def get_token_data():
    """토큰 파일에서 토큰 데이터를 가져옵니다."""
    token_path = os.path.join(config_dir, 'token.json')
    try:
        with open(token_path, 'r') as file:
            return json.load(file)
    except (FileNotFoundError, json.JSONDecodeError) as e:
        logging.error(f"토큰 파일 읽기 오류: {e}")
        return None

def save_token_data(token_data):
    """토큰 데이터를 토큰 파일에 저장합니다."""
    token_path = os.path.join(config_dir, 'token.json')
    try:
        os.makedirs(os.path.dirname(token_path), exist_ok=True)
        with open(token_path, 'w') as file:
            json.dump(token_data, file)
        logging.info(f"토큰이 {token_path}에 저장되었습니다.")
    except Exception as e:
        logging.error(f"토큰 저장 중 오류 발생: {e}")

def is_token_valid(token_data):
    """토큰의 만료 시간을 기준으로 유효성을 검사합니다."""
    TOKEN_REFRESH_THRESHOLD = timedelta(minutes=10)
    if not token_data or 'access_token' not in token_data:
        return False
    expiration_str = token_data.get('expiration')
    if not expiration_str:
        return False
    expiration = datetime.strptime(expiration_str, '%Y-%m-%d %H:%M:%S')
    return expiration > datetime.now() + TOKEN_REFRESH_THRESHOLD

def get_valid_token():
    """유효한 토큰을 가져오고, 필요시 갱신합니다."""
    token_data = get_token_data()
    if is_token_valid(token_data):
        logging.info("기존 토큰 사용 중")
        return token_data['access_token']
    else:
        logging.info("토큰이 만료되었거나 유효하지 않습니다. 갱신 중...")
        return refresh_token()

def refresh_token(retries=3, delay=60):
    """액세스 토큰을 갱신하며, 실패 시 재시도합니다."""
    url = "https://openapi.koreainvestment.com:9443/oauth2/tokenP"
    data = {
        'grant_type': 'client_credentials',
        'appkey': os.getenv('APPKEY'),
        'appsecret': os.getenv('APPSECRET')
    }
    headers = {'Content-Type': 'application/json'}
    
    for attempt in range(retries):
        try:
            logging.info("새 토큰 요청 중...")
            response = requests.post(url, headers=headers, json=data)
            if response.status_code == 200:
                response_data = response.json()
                new_token = response_data.get('access_token')
                expires_in = response_data.get('access_token_token_expired')

                try:
                    expiration = datetime.strptime(expires_in, '%Y-%m-%d %H:%M:%S')
                except ValueError:
                    expiration = datetime.now() + timedelta(seconds=int(expires_in) if expires_in.isdigit() else 3600)

                token_data = {'access_token': new_token, 'expiration': expiration.strftime('%Y-%m-%d %H:%M:%S')}
                save_token_data(token_data)
                logging.info(f"새 토큰 저장됨. 만료 시간: {expiration}")
                return new_token
            else:
                logging.error(f"토큰 갱신 오류: {response.text}")
                if "EGW00133" in response.text:
                    logging.info(f"{delay}초 후 재시도...")
                    time.sleep(delay)
                else:
                    return None
        except requests.exceptions.RequestException as e:
            logging.error(f"토큰 갱신 중 요청 예외 발생: {e}")
            return None
    return None

# data_dir를 정의
data_dir = os.path.join(project_root, 'data')

# main_total 모듈 초기화 및 실행
import main_total
main_total.initialize()
main_total.run_main()


def get_api_response(url, headers, params):
    """Send a GET request to the specified URL with headers and parameters."""
    try:
        response = requests.get(url, headers=headers, params=params)
        response.raise_for_status()
        return response.json()
    except requests.exceptions.RequestException as e:
        logging.error(f"Request exception occurred: {e}")
        return None

def get_search_conditions():
    """Retrieve search conditions from the API."""
    url = "https://openapi.koreainvestment.com:9443/uapi/domestic-stock/v1/quotations/psearch-title"
    headers = {
        'Content-Type': 'application/json; charset=utf-8',
        'Authorization': f'Bearer {get_valid_token()}',
        'appkey': os.getenv('APPKEY'),
        'appsecret': os.getenv('APPSECRET'),
        'tr_id': 'HHKST03900300'
    }
    params = {'user_id': os.getenv('USER_ID')}
    logging.info(f"Request URL: {url}")
    logging.info(f"Request Headers: {headers}")
    logging.info(f"Request Params: {params}")

    response = get_api_response(url, headers, params)
    logging.info(f"API Response: {response}")

    if response and 'output2' in response:
        return response['output2']
    else:
        logging.error(f"Unexpected response: {response}")
        return None

def get_search_results(seq):
    """Retrieve search results from the API using the provided sequence."""
    url = "https://openapi.koreainvestment.com:9443/uapi/domestic-stock/v1/quotations/psearch-result"
    headers = {
        'Content-Type': 'application/json; charset=utf-8',
        'Authorization': f'Bearer {get_valid_token()}',
        'appkey': os.getenv('APPKEY'),
        'appsecret': os.getenv('APPSECRET'),
        'tr_id': 'HHKST03900400'
    }
    params = {'user_id': os.getenv('USER_ID'), 'seq': seq}

    logging.info(f"Request URL: {url}")
    logging.info(f"Request Headers: {headers}")
    logging.info(f"Request Params: {params}")

    response = get_api_response(url, headers, params)
    logging.info(f"API Response: {response}")

    if response and 'output2' in response:
        return response['output2']
    else:
        logging.error(f"Unexpected response: {response}")
        return None

def get_stock_codes_and_names(seq):
    """Retrieve stock codes and names from search results."""
    search_conditions = get_search_conditions()
    if not search_conditions:
        logging.error("No search conditions found.")
        return []

    logging.info(f"Search conditions seq: {seq}")
    search_results = get_search_results(seq)
    if not search_results:
        logging.error("No search results found.")
        return []

    stock_codes_and_names = []
    for result in search_results:
        code = result.get('code')
        name = result.get('name')
        logging.info(f"Retrieved code: {code}, name: {name}")
        if code and name:
            stock_codes_and_names.append((code.zfill(6), name))
        else:
            logging.error(f"Invalid stock code or name: {result}")

    logging.info(f"Detected stock codes and names: {stock_codes_and_names}")
    save_stock_codes_to_csv(stock_codes_and_names)
    return stock_codes_and_names

def get_stock_codes_from_csv(file_path):
    """Read stock codes from the CSV file."""
    try:
        df = pd.read_csv(file_path, dtype={'code': str}, usecols=['code'])
        if df.empty:
            raise ValueError("The CSV file is empty.")
        return df['code'].tolist()
    except FileNotFoundError:
        logging.error(f"CSV file not found: {file_path}")
        return []
    except pd.errors.EmptyDataError:
        logging.error(f"CSV file is empty: {file_path}")
        return []
    except pd.errors.ParserError:
        logging.error(f"Error parsing CSV file: {file_path}")
        return []
    except Exception as e:
        logging.error(f"Error reading stock codes from CSV file: {e}")
        return []


def read_stock_data(stock_code):
    try:
        directory_path = os.path.join(data_dir, 'stock_data')
        dataframes = []
        for file_name in os.listdir(directory_path):
            if file_name.startswith(stock_code) and file_name.endswith('_data.csv'):
                file_path = os.path.join(directory_path, file_name)
                logging.info(f"Reading data from {file_path}")
                df = pd.read_csv(file_path)
                
                if df is None or df.empty:
                    logging.error(f"No data in {file_path} for stock code {stock_code}")
                    continue

                # 필수 열이 존재하는지 확인
                required_columns = ['open', 'high', 'low', 'close', 'volume', 'date']
                missing_columns = [col for col in required_columns if col not in df.columns]
                if missing_columns:
                    logging.error(f"Missing columns: {missing_columns} in {file_path}")
                    continue  # 현재 파일을 건너뛰고 다음 파일로 진행

                # 날짜 열을 올바르게 변환
                try:
                    df['date'] = pd.to_datetime(df['date'], errors='coerce')  # 날짜 형식으로 변환
                except Exception as e:
                    logging.error(f"Error converting 'date' column to datetime in {file_path}: {e}")
                    continue

                # 숫자 형식으로 변환 가능한 열 처리
                try:
                    for col in ['open', 'high', 'low', 'close', 'volume']:
                        df[col] = pd.to_numeric(df[col], errors='coerce')

                    # NaN 값을 제거하거나 0으로 대체
                    df.dropna(subset=['open', 'high', 'low', 'close', 'volume', 'date'], inplace=True)

                    # 모든 값이 0이거나 None인 행 제거
                    df = df[(df[['open', 'high', 'low', 'close', 'volume']] != 0).any(axis=1)]

                    # 숫자 열을 float으로 유지
                    df[['open', 'high', 'low', 'close', 'volume']] = df[['open', 'high', 'low', 'close', 'volume']].astype(float)

                    # 비어 있는 행을 다음 행의 데이터로 채우기
                    df.ffill(inplace=True)
                except (ValueError, TypeError) as ve:
                    logging.error(f"Error converting columns to numeric in {file_path}: {ve}")
                    continue  # 현재 파일을 건너뛰고 다음 파일로 진행

                dataframes.append(df)
        
        if dataframes:
            return pd.concat(dataframes, ignore_index=True)
        else:
            logging.error(f"No valid data files found for stock {stock_code} in directory: {directory_path}")
            return pd.DataFrame()
    except pd.errors.EmptyDataError as e:
        logging.error(f"Empty data error for {stock_code}: {e}")
        return pd.DataFrame()
    except pd.errors.ParserError as e:
        logging.error(f"Parsing error for {stock_code}: {e}")
        return pd.DataFrame()
    except Exception as e:
        logging.error(f"Error reading stock data for {stock_code}: {e}")
        return pd.DataFrame()



def save_stock_codes_to_csv(stock_codes_and_names):
    """Save stock codes and names to a CSV file."""
    file_path = os.path.join(data_dir, 'stock_codes.csv')
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    
    if os.path.exists(file_path):
        existing_df = pd.read_csv(file_path)
    else:
        existing_df = pd.DataFrame(columns=['code', 'name'])
    
    # Ensure stock codes are 6 characters long, padded with leading zeros if necessary
    stock_codes_and_names = [(code.zfill(6), name) for code, name in stock_codes_and_names]
    
    new_df = pd.DataFrame(stock_codes_and_names, columns=['code', 'name'])
    combined_df = pd.concat([existing_df, new_df]).drop_duplicates().reset_index(drop=True)
    
    combined_df.to_csv(file_path, index=False)
    logging.info(f"Stock codes saved to {file_path}")


def another_function():
    # 다른 함수에서도 json을 사용
    try:
        data = '{"name": "John", "age": 30}'
        parsed_data = json.loads(data)
    except Exception as e:
        logging.error(f"An error occurred: {e}")


def update_data():
    global seq
    stock_codes_and_names = get_stock_codes_and_names(seq)
    if not stock_codes_and_names:
        logging.error("No stock codes detected.")
        seq = 1 if seq == 0 else 0  # Toggle seq between 0 and 1
        stock_codes_and_names = get_stock_codes_and_names(seq)
        if not stock_codes_and_names:
            logging.error("No stock codes detected after second attempt. Using default stock codes.")
            stock_codes_and_names = [
                ('005930', 'Samsung Electronics'),
                ('000660', 'SK Hynix'),
                ('035420', 'NAVER')
            ]

    fetch_data_for_detected_stocks()

    stock_data = []
    for code, name in stock_codes_and_names:
        stock_df = read_stock_data(code)
        if stock_df is not None and not stock_df.empty:
            stock_data.append(stock_df)

    if not stock_data:
        logging.error("No stock data available.")
        return None, None, None, None, None

    stock_data = pd.concat(stock_data, ignore_index=True)
    stock_data = add_features(stock_data)

    financial_data = load_csv_files_from_directory(os.path.join(data_dir, 'financial_data'))
    news_data = load_csv_files_from_directory(os.path.join(data_dir, 'news_data'))
    economic_data = load_csv_files_from_directory(os.path.join(data_dir, 'economic_data'))
    return stock_data, financial_data, news_data, economic_data, stock_codes_and_names[0][0]  # 첫 번째 주식 코드 반환

def preprocess_and_save_data():
    """주식 데이터와 전체 데이터를 각각 전처리하고 저장."""
    # 데이터 준비 및 전처리
    stock_data, financial_data, news_data, economic_data, detected_stock_code = update_data()
    if stock_data is None:
        logging.error("Data update failed.")
        return
    
    # 주식 데이터만 처리하여 별도로 저장
    stock_preprocessed = preprocess_data(stock_data, None, None, None)
    stock_preprocessed.to_csv(os.path.join(data_dir, 'preprocessed_data_stock.csv'), index=False)

    # 전체 데이터 처리 및 저장
    full_preprocessed = preprocess_data(stock_data, financial_data, news_data, economic_data)
    full_preprocessed.to_csv(os.path.join(data_dir, 'preprocessed_data.csv'), index=False)


def integrate_news_data(stock_data, news_data):
    required_columns = ['date', 'headline', 'content']  # 필요에 따라 변경
    missing_columns = [col for col in required_columns if col not in news_data.columns]
    if missing_columns:
        logging.error(f"Missing required columns in news data: {missing_columns}")
        return stock_data

    news_data['date'] = pd.to_datetime(news_data['date'])
    stock_data['date'] = pd.to_datetime(stock_data['date'])
    return stock_data.merge(news_data, how='left', left_on='date', right_on='date')

def normalize_data(df):
    """Normalize the data using StandardScaler."""
    scaler = StandardScaler()
    normalized_df = pd.DataFrame(scaler.fit_transform(df), columns=df.columns)
    return normalized_df, scaler

def run_trading_during_market_hours(stock_model, stock_scaler):
    # 거래 시간 동안 실시간 데이터 수집 및 예측 수행
    logging.info("Market is open. Running real-time data collection and prediction.")
    
    stock_codes_path = os.path.join(data_dir, 'stock_codes.csv')
    stock_codes = get_stock_codes_from_csv(stock_codes_path)

    all_predictions = {}  # 모든 종목의 예측을 저장할 딕셔너리

    for ticker in stock_codes:
        # 실시간 예측 및 모델 업데이트 (주식 데이터 모델 사용)
        real_time_predictions = real_time_prediction(stock_model, stock_scaler, ticker)
        if real_time_predictions is not None:
            logging.info(f"Real-time predictions for {ticker}: {real_time_predictions}")
            all_predictions[ticker] = real_time_predictions  # 예측을 딕셔너리에 저장
        else:
            logging.error(f"Failed to fetch real-time predictions for {ticker}")

        updated_model = update_real_time_model(stock_model, stock_scaler, ticker)
        if isinstance(updated_model, tuple):
            updated_model = updated_model[0]  # 필요한 객체를 선택
        save_model(updated_model, os.path.join(data_dir, 'models', 'updated_lstm_model_stock.h5'), stock_model_json_path, stock_scaler)
        
        # 요청 간격을 0.05초로 설정
        time.sleep(0.05)

    execute_trades(all_predictions)  # 모든 예측을 사용하여 거래 실행

def load_and_validate_data(file_path):
    """파일을 로드하고 데이터프레임을 반환하며, 데이터가 없거나 오류가 발생하면 None을 반환합니다."""
    try:
        df = pd.read_csv(file_path)
        if df is None or df.empty:
            logging.error(f"Data file is empty or could not be loaded: {file_path}")
            return None
        logging.info(f"First few rows of {file_path}:\n{df.head()}")
        return df
    except FileNotFoundError:
        logging.error(f"Data file not found: {file_path}")
        return None
    except pd.errors.EmptyDataError:
        logging.error(f"Data file is empty: {file_path}")
        return None
    except pd.errors.ParserError:
        logging.error(f"Parsing error occurred while reading {file_path}")
        return None
    except Exception as e:
        logging.error(f"An error occurred while loading {file_path}: {e}")
        return None
# main_total.py

def run_main():
    logging.info("Running main function of main_total module...")
    # 여기에 실제 메인 로직을 추가합니다.

def main():
    initialize()
    run_main()

    # stock_data_file_path 변수를 여기서 명확하게 정의합니다.
    stock_data_file_path = os.path.join(data_dir, 'stock_data.csv')

    if not os.path.exists(stock_data_file_path):
        logging.error(f"File not found: {stock_data_file_path}")
        return

    stock_codes = get_stock_codes_from_csv(stock_data_file_path)
    if stock_codes:
        logging.info(f"Loaded stock codes: {stock_codes}")
    else:
        logging.error("No valid stock data to process.")

    market_open_time = dt_time(9, 0)
    market_close_time = dt_time(15, 0)
    current_time = datetime.now().time()

    stock_model = None
    stock_scaler = None

    if os.path.exists(stock_model_path) and os.path.exists(stock_model_json_path) and os.path.exists(stock_scaler_path):
        logging.info("Loading existing stock model and scaler.")
        stock_model = load_model(stock_model_path)
        stock_scaler = joblib.load(stock_scaler_path)
    else:
        logging.warning("No existing stock model found. Checking for preprocessed data.")
        if not os.path.exists(os.path.join(data_dir, 'preprocessed_data_stock.csv')):
            logging.info("Preprocessed data not found. Running preprocessing.")
            preprocess_and_save_data()

        if os.path.exists(os.path.join(data_dir, 'preprocessed_data_stock.csv')):
            stock_data = pd.read_csv(os.path.join(data_dir, 'preprocessed_data_stock.csv'))
            (X_train, X_test, y_train, y_test), stock_scaler, _ = preprocess_data(stock_data, None, None, None)
            stock_model = train_lstm_model(X_train, y_train)
            save_model(stock_model, stock_model_path, stock_model_json_path, stock_scaler)
        else:
            logging.error("Preprocessed data still not found. Cannot proceed with model training.")

    if current_time >= market_close_time or current_time < market_open_time:
        try:
            preprocess_and_save_data()

            if os.path.exists(all_model_path) and os.path.exists(all_model_json_path) and os.path.exists(all_scaler_path):
                logging.info("Loading existing all data model and scaler.")
                all_model = load_model(all_model_path)
                all_scaler = joblib.load(all_scaler_path)
            else:
                logging.info("No existing all data model found. Training new all data model.")
                all_data = pd.read_csv(os.path.join(data_dir, 'preprocessed_data.csv'))
                (X_train, X_test, y_train, y_test), all_scaler, _ = preprocess_data(all_data, None, None, None)
                all_model = train_lstm_model(X_train, y_train)
                save_model(all_model, all_model_path, all_model_json_path, all_scaler)
        except Exception as e:
            logging.error(f"An error occurred: {e}")
    else:
        logging.info("Current time is within market hours. Running real-time trading tasks.")
        if stock_model and stock_scaler:
            run_trading_during_market_hours(stock_model, stock_scaler)
        else:
            logging.error("Stock model or scaler not loaded. Cannot perform real-time trading.")

if __name__ == "__main__":
    main()