import requests
from textblob import TextBlob

class NewsSentimentAnalyzer:
    def __init__(self, api_key):
        self.api_key = api_key

    def fetch_news(self, query):
        url = f"https://newsapi.org/v2/everything?q={query}&apiKey={self.api_key}"
        response = requests.get(url)
        news_data = response.json()
        return news_data

    def analyze_sentiment(self, news_data):
        sentiments = []
        for article in news_data['articles']:
            text = article['title'] + ". " + article['description']
            sentiment = TextBlob(text).sentiment.polarity
            sentiments.append(sentiment)
        average_sentiment = sum(sentiments) / len(sentiments) if sentiments else 0
        return average_sentiment

    def get_stock_sentiment(self, stock_code):
        news_data = self.fetch_news(stock_code)
        sentiment = self.analyze_sentiment(news_data)
        return sentiment
